==> zim_0.23.orig.tar.gz: Zim-0.23/lib/Zim/Formats/Wiki.pm <==
package Zim::Formats::Wiki;

use strict;
no warnings;
use Zim::Formats;

our $VERSION = '0.21';
our @ISA = qw/Zim::Formats/;

# TODO: some tags can be nested: email links for example

=head1 NAME

Zim::Formats::Wiki - Wiki text parser

=head1 DESCRIPTION

This is the default parser for Zim.
It uses a wiki-style syntax to format plain text.

FIXME more verbose description

All format types are signified by double characters,
this is done to prevent accidental formatting when the same
characters are used normally in a text. For the same reason
the strike character is '~' instead of '-' because '--' can
occur in normal ascii text. In the regexes we try to always
match the inner pair if more than two of these characters
are encountered. (Thus if we see C<***bold***> we should match
C<*(**bold**)*>.)


=head1 METHODS

=over 4

=item C<load_tree(IO, PAGE)>

Reads plain text from a filehandle and returns a parse tree.

=cut

sub load_tree { # TODO whitelines between verbatim blocks should be preserved
	my ($class, $io, $page) = @_;

	my @tree;
	my $para = '';
	while (<$io>) {
		s/\r?\n$/\n/; # DOS to Unix conversion
		if (!/\S/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			$tree[-1][1]{empty_lines}++ if scalar @tree;
			$para = '';
		}
		elsif (/^==+\s+\S+/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			push @tree, $class->parse_head($_);
			$para = '';
		}
		#elsif (/^----+\s*$/) {
		#	push @tree, $class->parse_para($para) if $para =~ /\S/;
		#	push @tree, ['hr', {} ];
		#	$para = '';
		#}
		else { $para .= $_ }
	}
	push @tree, $class->parse_para($para) if $para =~ /\S/;
	
	#use Data::Dumper; print STDERR Dumper \@tree;
	return ['Page', {}, @tree];
}

sub parse_head { # parse a header
	my ($class, $head) = @_;
	chomp $head;
	$head =~ s/^(==+)\s+(.*?)(\s+==+\s*|\s*)$/$2/;
	my $level = 7 - length($1); # =X6 => head1, =X5 => head2 etc.
	$level = 1 if $level < 1;
	return ['head'.$level, {}, $head];
}

sub parse_para { # distinguise normal paras from verbatim paras
	my ($class, $text) = @_;
	return  $class->parse_Verbatim($text) unless $text =~ /^(?!\t|\s\s+)/m;
	#$text =~ s/^\.//mg unless $text =~ /^(?!\.?\t|\.?\s\s+)/s;
	$text =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg;
	return ['Para', {}, $class->parse_block($text)];
}

sub parse_Verbatim { # parse pre formated paragraphs
	my ($class, $pre) = @_;
	my ($indent) = ($pre =~ /^(\s+)/);
	$pre =~ s/^$indent//mg;
	return ['Verbatim', {}, $pre];
}

our @parser_subs = qw/
	parse_verbatim
	parse_links
	parse_images
	parse_styles
	parse_urls
/;

sub parse_block { # parse a block of text
	my ($class, @text) = @_;
	for my $sub (@parser_subs) {
		@text = map {ref($_) ? $_ : ($class->$sub($_))} @text;
	}
	return @text;
}

sub parse_verbatim { # like parse_style() but needs to be done earlier
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\'\'(.+)\'\'$/) { ['verbatim',  {}, $1] }
	} split /(\'\'(?!\').+?\'\')/, $text;
}

sub parse_links {
	my ($class, $text) = @_;
	my $i = 0;
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(\[\[(?!\[).+?\]\])#, $text;
}

sub _parse_link {
	my $l = shift;
	my $text;
	if ($l =~ /^\[\[([^|]*)\|?(.*)\]\]$/) { # [[link]] or [[link|text]]
		($l, $text) = ($1, $2);
		$l = $text unless length $l; # [[|link]] bug
	}
	if ($l =~ /^mailto:|^\S+\@\S+\.\w+$/) {
		$text = $l unless length $text;
		$l =~ s/^(mailto:)?/mailto:/;
	}
	return ['link', {to => $l}, length($text) ? $text : $l ];
}

sub parse_urls {
	my ($class, $text) = @_;
	my $i = 0;
	my $C = q/[^\s\"\<\>\']/; # limit the character class a bit
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(
			\b\w[\w\+\-\.]+://  $C*\[$C+\](?:$C+[\w\/])? |
			\b\w[\w\+\-\.]+://  $C+[\w\/]                |
			\bmailto:$C+\@      $C*\[$C+\](?:$C+[\w/])?  |
			\bmailto:$C+\@      $C+[\w/]                 |
			\b$C+\@$C+\.\w+\b
		)#x, $text;
		# The host name in an uri can be "[hex:hex:..]" for ipv6
		# but we do not want to match "[http://foo.org]"
		# See rfc3986 for the official -but unpractical- regex
}

sub parse_styles { # parse blocks of bold, italic and underline
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\*\*(.+)\*\*$/) { ['bold',      {}, $1] }
		elsif (/^\/\/(.+)\/\/$/) { ['italic',    {}, $1] }
		elsif (/^\~\~(.+)\~\~$/) { ['strike',    {}, $1] }
		elsif (/^\_\_(.+)\_\_$/) { ['underline', {}, $1] }
	} split /(
		\*\*(?!\*).+?\*\* |
		(?<!\:)\/\/(?!\/).+?\/\/ |
		\~\~(?!\~).+?\~\~ |
		__(?!_).+?__
	)/x, $text;
}

sub parse_images {
	my ($class, $text) = @_;
	my $i = 0;
	my @parts =
		map   { ($i++ % 2) ? ['image', {src => $_}] : $_ }
		split /\{\{(?!\{)(.+?)\}\}/, $text;
	for my $p (@parts) {
		next unless ref $p and $$p[0] eq 'image';
		if ($$p[1]{src} =~ s/\?(\w+=\w+(?:&\w+=\w+)*)$//) {
			my %arg = split /[=&]/, $1;
			$$p[1]{$_} = $arg{$_} for keys %arg;
		}
	}
	return @parts;
}

=item C<save_tree(IO, TREE, PAGE)>

Serializes the parse tree into a piece of plain text and writes this
to a filehandle.

=cut

sub save_tree {
	# TODO add support for recursive tags
	my ($class, $io, $tree) = @_;

	my $old_fh = select $io;
	eval { $class->_save_tree($tree) };
	select $old_fh;
	die $@ if $@;
}

sub _save_tree {
	my ($class, $tree) = @_;
	
	my ($name, $opt) = splice @$tree, 0, 2;
	die "Invalid parse tree"
		unless length $name and ref($opt) eq 'HASH';

	while (@$tree) {
		my $node = shift @$tree;
		unless (ref $node) {
			$node =~ s/^(\s*)\x{2022}(\s+)/$1*$2/mg;
			print $node;
			next;
		}
	
		my ($tag, $meta) = @$node[0, 1];
		if ($tag eq 'Para') {
			$class->_save_tree($node);
			print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
			next;
		}
		
		my $text = _dump($node);		
		if ($tag eq 'Verbatim')  {
			$text =~ s/^/\t/mg;
			print $text;
		}
		elsif ($tag =~ /^head(\d)$/) {
			my $n = 7 - $1;
			print ( ('='x$n)." $text ".('='x$n)."\n" );
		}
		elsif ($tag eq 'image') {
			my $file = $$meta{src};
			my @k = sort grep {$_ ne 'src'} keys %$meta;
			$file .= '?' . join '&', map "$_=$$meta{$_}", @k if @k;
			print '{{'.$file.'}}';
		}
		elsif ($tag eq 'link') {
			my $to = $$meta{to};
			$to =~ s/^mailto:// unless $text =~ /^mailto:/;
			print (
				($to ne $text) ? "[[$to|$text]]" :
				($to !~ /\s/ and $to =~ m#^\w[\w\+\-\.]+://|^(?:mailto:)?\S+\@\S+\.\w+#)
					? $to : "[[$to]]" );
		}
		# per line markup for remaining tags ...
		elsif ($tag eq 'bold') {
			print map { /\S/ ? "**$_**" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'italic') {
			print map { /\S/ ? "//$_//" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'strike') {
			print map { /\S/ ? "~~$_~~" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'underline') {
			print map { /\S/ ? "__$_\__" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'verbatim') {
			print map { /\S/ ? "''$_''" : $_} split /(\n)/, $text
		}
		else { die "Unkown tag in wiki parse tree: $tag\n" }
		
		print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
	}
}

sub _dump {
	my $node = shift;
	splice @$node, 0, 2;
	for (@$node) { $_ = _dump($_) if ref $_ }
	return join '', @$node;
}

1;

__END__

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>,
L<Zim::Page>

=cut

==> zim_0.25.orig.tar.gz: Zim-0.25/lib/Zim/Formats/Wiki.pm <==
package Zim::Formats::Wiki;

use strict;
no warnings;
use Zim::Formats;

our $VERSION = '0.24';
our @ISA = qw/Zim::Formats/;

# TODO: some tags can be nested: email links for example

=head1 NAME

Zim::Formats::Wiki - Wiki text parser

=head1 DESCRIPTION

This is the default parser for Zim.
It uses a wiki-style syntax to format plain text.

FIXME more verbose description

All format types are signified by double characters,
this is done to prevent accidental formatting when the same
characters are used normally in a text. For the same reason
the strike character is '~' instead of '-' because '--' can
occur in normal ascii text. In the regexes we try to always
match the inner pair if more than two of these characters
are encountered. (Thus if we see C<***bold***> we should match
C<*(**bold**)*>.)


=head1 METHODS

=over 4

=item C<load_tree(IO, PAGE)>

Reads plain text from a filehandle and returns a parse tree.

=cut

sub load_tree { # TODO whitelines between verbatim blocks should be preserved
	my ($class, $io, $page) = @_;

	my @tree;
	my $para = '';
	while (<$io>) {
		s/\r?\n$/\n/; # DOS to Unix conversion
		if (!/\S/) {
			push @tree, $class->parse_para($para, $page) if $para =~ /\S/;
			$tree[-1][1]{empty_lines}++ if scalar @tree;
			$para = '';
		}
		elsif (/^==+\s+\S+/) {
			push @tree, $class->parse_para($para, $page) if $para =~ /\S/;
			push @tree, $class->parse_head($_);
			$para = '';
		}
		#elsif (/^----+\s*$/) {
		#	push @tree, $class->parse_para($para, $page) if $para =~ /\S/;
		#	push @tree, ['hr', {} ];
		#	$para = '';
		#}
		else { $para .= $_ }
	}
	push @tree, $class->parse_para($para, $page) if $para =~ /\S/;
	
	#use Data::Dumper; print STDERR Dumper \@tree;
	return ['Page', {}, @tree];
}

sub parse_head { # parse a header
	my ($class, $head) = @_;
	chomp $head;
	$head =~ s/^(==+)\s+(.*?)(\s+==+\s*|\s*)$/$2/;
	my $level = 7 - length($1); # =X6 => head1, =X5 => head2 etc.
	$level = 1 if $level < 1;
	return ['head'.$level, {}, $head];
}

sub parse_para { # distinguise normal paras from verbatim paras
	my ($class, $text, $page) = @_;
	return  $class->parse_Verbatim($text) unless $text =~ /^(?!\t|\s\s+)/m;
	#$text =~ s/^\.//mg unless $text =~ /^(?!\.?\t|\.?\s\s+)/s;
	$text =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg;
	return ['Para', {}, $class->parse_block($text, $page)];
}

sub parse_Verbatim { # parse pre formated paragraphs
	my ($class, $pre) = @_;
	my ($indent) = ($pre =~ /^(\s+)/);
	$pre =~ s/^$indent//mg;
	return ['Verbatim', {}, $pre];
}

our @parser_subs = qw/
	parse_verbatim
	parse_links
	parse_images
	parse_styles
	parse_urls
/;

sub parse_block { # parse a block of text
	my ($class, $text, $page) = @_;
	my @text = ($text);
	for my $sub (@parser_subs) {
		@text = map {ref($_) ? $_ : ($class->$sub($_, $page))} @text;
	}
	return @text;
}

sub parse_verbatim { # like parse_style() but needs to be done earlier
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\'\'(.+)\'\'$/) { ['verbatim',  {}, $1] }
	} split /(\'\'(?!\').+?\'\')/, $text;
}

sub parse_links {
	my ($class, $text) = @_;
	my $i = 0;
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(\[\[(?!\[).+?\]\])#, $text;
}

sub _parse_link {
	my $l = shift;
	my $text;
	if ($l =~ /^\[\[([^|]*)\|?(.*)\]\]$/) { # [[link]] or [[link|text]]
		($l, $text) = ($1, $2);
		$l = $text unless length $l; # [[|link]] bug
	}
	if ($l =~ /^mailto:|^\S+\@\S+\.\w+$/) {
		$text = $l unless length $text;
		$l =~ s/^(mailto:)?/mailto:/;
	}
	return ['link', {to => $l}, length($text) ? $text : $l ];
}

sub parse_urls {
	my ($class, $text) = @_;
	my $i = 0;
	my $C = q/[^\s\"\<\>\']/; # limit the character class a bit
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(
			\b\w[\w\+\-\.]+://  $C*\[$C+\](?:$C+[\w\/])? |
			\b\w[\w\+\-\.]+://  $C+[\w\/]                |
			\bmailto:$C+\@      $C*\[$C+\](?:$C+[\w/])?  |
			\bmailto:$C+\@      $C+[\w/]                 |
			\b$C+\@$C+\.\w+\b
		)#x, $text;
		# The host name in an uri can be "[hex:hex:..]" for ipv6
		# but we do not want to match "[http://foo.org]"
		# See rfc/3986 for the official -but unpractical- regex
}

sub parse_styles { # parse blocks of bold, italic and underline
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\*\*(.+)\*\*$/) { ['bold',      {}, $1] }
		elsif (/^\/\/(.+)\/\/$/) { ['italic',    {}, $1] }
		elsif (/^\~\~(.+)\~\~$/) { ['strike',    {}, $1] }
		elsif (/^\_\_(.+)\_\_$/) { ['underline', {}, $1] }
	} split /(
		\*\*(?!\*).+?\*\* |
		(?<!\:)\/\/(?!\/).+?\/\/ |
		\~\~(?!\~).+?\~\~ |
		__(?!_).+?__
	)/x, $text;
}

sub parse_images {
	my ($class, $text, $page) = @_;
	my $i = 0;
	my @parts =
		map   { ($i++ % 2) ? ['image', {src => $_}] : $_ }
		split /\{\{(?!\{)(.+?)\}\}/, $text;
	for my $p (@parts) {
		next unless ref $p and $$p[0] eq 'image';
		if ($$p[1]{src} =~ s/\?(\w+=\w+(?:&\w+=\w+)*)$//) {
			my %arg = split /[=&]/, $1;
			$$p[1]{$_} = $arg{$_} for keys %arg;
		}
		$$p[1]{file} = $page->resolve_file($$p[1]{src});
		#use Data::Dumper; warn "IMAGE: ", Dumper $p;
	}
	return @parts;
}

=item C<save_tree(IO, TREE, PAGE)>

Serializes the parse tree into a piece of plain text and writes this
to a filehandle.

=cut

sub save_tree {
	# TODO add support for recursive tags
	my ($class, $io, $tree) = @_;

	my $old_fh = select $io;
	eval { $class->_save_tree($tree) };
	select $old_fh;
	die $@ if $@;
}

sub _save_tree {
	my ($class, $tree) = @_;
	
	my ($name, $opt) = splice @$tree, 0, 2;
	die "Invalid parse tree"
		unless length $name and ref($opt) eq 'HASH';

	while (@$tree) {
		my $node = shift @$tree;
		unless (ref $node) {
			$node =~ s/^(\s*)\x{2022}(\s+)/$1*$2/mg;
			print $node;
			next;
		}
	
		my ($tag, $meta) = @$node[0, 1];
		if ($tag eq 'Para') {
			$class->_save_tree($node);
			print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
			next;
		}
		
		my $text = _dump($node);		
		if ($tag eq 'Verbatim')  {
			$text =~ s/^/\t/mg;
			print $text;
		}
		elsif ($tag =~ /^head(\d)$/) {
			my $n = 7 - $1;
			print ( ('='x$n)." $text ".('='x$n)."\n" );
		}
		elsif ($tag eq 'image') {
			my $file = $$meta{src};
			my @k = sort grep {$_ !~ /^(src|file)$/} keys %$meta;
			$file .= '?' . join '&', map "$_=$$meta{$_}", @k if @k;
			print '{{'.$file.'}}';
		}
		elsif ($tag eq 'link') {
			my $to = $$meta{to};
			$to =~ s/^mailto:// unless $text =~ /^mailto:/;
			print (
				($to ne $text) ? "[[$to|$text]]" :
				($to !~ /\s/ and $to =~ m#^\w[\w\+\-\.]+://|^(?:mailto:)?\S+\@\S+\.\w+#)
					? $to : "[[$to]]" );
		}
		# per line markup for remaining tags ...
		elsif ($tag eq 'bold') {
			print map { /\S/ ? "**$_**" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'italic') {
			print map { /\S/ ? "//$_//" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'strike') {
			print map { /\S/ ? "~~$_~~" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'underline') {
			print map { /\S/ ? "__$_\__" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'verbatim') {
			print map { /\S/ ? "''$_''" : $_} split /(\n)/, $text
		}
		else { die "Unkown tag in wiki parse tree: $tag\n" }
		
		print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
	}
}

sub _dump {
	my $node = shift;
	splice @$node, 0, 2;
	for (@$node) { $_ = _dump($_) if ref $_ }
	return join '', @$node;
}

1;

__END__

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>,
L<Zim::Page>

=cut
